// Main.cpp
//Main flow to get faces from a graph-like representation of a polyhedrical shape

//Based on:
	//Varley P.A.C. and Company P.
	//A new algorithm for finding faces in wireframes.
	//Computer-Aided Design (ISSN 0010-4485). Volume 42, Issue 4, April 2010, Pages 279-309
//

#include "stdafx.h"
#include <vector>
#include <iostream>
#include <string>
#include <fstream>

#include "stdio.h"

#include "CueFacesV_Geometry.h"
#include "CueFacesVarley.h"


int _tmain(int argc, _TCHAR* argv[])
{
	//Input data
	//----------
	int VertexCount= 0;
	std::vector <double> VertexX;
	std::vector <double> VertexY;

	int EdgeCount= 0;
	std::vector <int> EdgeU;
	std::vector <int> EdgeV;

	std::string FileName;
	std::cout << "\nPlease enter your desired filename \n (extension .Sketch will be added automatically)\n" << std::endl;
	std::cin >> FileName;

	FileName += ".Sketch";
	std::ifstream in (FileName.c_str());

	//Reads Vertices and Edges
	if (in.is_open()){

		std::string line;
		while (getline(in, line)){
			if(line.substr(0, 6)== "Vertex"){
				long BeginX= line.find(':')+1;
				long EndX= line.find(',', BeginX);
				VertexX.push_back(atof(line.substr(BeginX, EndX-BeginX).c_str()));

				long BeginY= EndX+1;
				while(line.substr(BeginY, 1) == " ")
					BeginY++;
				long EndY= line.find(' ', BeginY);
				if(EndY<0)
					EndY= line.size()-1;
				VertexY.push_back(-atof(line.substr(BeginY, EndY-BeginY).c_str()));

				VertexCount++;
			}
			else if(line.substr(0, 4)== "Line"){
				EdgeU.push_back(0);
				long BeginV1 = line.find('V')+1;
				long EndV1 = line.find(' ', BeginV1);
				sscanf_s(line.substr(BeginV1, EndV1-BeginV1).c_str(), "%d", &EdgeU[EdgeCount]);

				EdgeV.push_back(0);
				long BeginV2 = line.rfind('V')+1;
				long EndV2 = line.size();
				sscanf_s(line.substr(BeginV2, EndV2-BeginV2).c_str(), "%d", &EdgeV[EdgeCount]);
				EdgeCount++;
			}
		} //while

		// Closes the file
		in.close();
	}
	else
		printf("Unable to open file\n");

	//Find faces
	//----------
	CFacesVarley FVarley(VertexCount, VertexX, VertexY,
						 EdgeCount, EdgeU, EdgeV);
	std::vector <FACE> Face= FVarley.Get_Faces_Varley();

	//Print output
	//------------
	printf("Example %s \n",FileName.c_str());

	int FaceCount= Face.size();
	printf("Number of faces= %3d \n", FaceCount);

	for(long i= 0; i<FaceCount; i++){
		printf("Face number %3d \n", i);
		for(long j= 0; j<Face[i].C; j++){
			printf("Edge %3d \n", Face[i].E[j]);
		}
		for(long j= 0; j<(long)Face[i].V.size(); j++){
			printf("Vertex %3d \n", Face[i].V[j]);
		}
	}

	printf("\n");
	system("PAUSE");

	return 0;
}
